def _print_line_informations(choices=None, default=None):
    if choices is not None:
        all_choices = [
            str(chc).upper() if chc == default else str(chc) for chc in choices
        ]
        print(f"[{'/'.join(all_choices)}]", end=" ")
    if default not in {"", None}:
        if choices and default not in choices:
            for new_default, value in choices.items():
                if value == default:
                    default = new_default
                    break
        print(f"(default: {default!r})", end=" ")
    return default


def ask(text, choices=None, default=None, required=False, cast=str):
    choices = [] if choices is None else [cast(chc) for chc in choices]
    while True:
        print(text, end=" ")
        default = _print_line_informations(choices, default)
        print(":", end=" ")

        result = input()

        if cast is not None:
            try:
                result = cast(result)
            except ValueError:
                print("Couldn't cast %r to %r" % (result, cast))
                continue

        if not result and required:
            print("you must provide an answer")
        elif result and choices and result not in choices:
            print(f"{result!r} is not in {choices!r}")
        else:
            return result


def ask_bool(text, default=True, required=False):
    default = "yes" if default else "no"
    result = ask(
        text,
        choices=["y", "yes", "n", "no"],
        default=default,
        required=required,
        cast=str,
    )
    if result.lower() in {"y", "yes"}:
        return True
    return False



import logging
from copy import deepcopy
from itertools import chain

from the_conf.utils import TYPE_MAPPING, Index, NoValue
from the_conf.files import extract_value

logger = logging.getLogger(__name__)


class AbstractNode:
    def __init__(self, parameters=None, parent=None, name=""):
        self._name = name
        self._parent = parent
        self._parameters = {}
        self._children = []
        self._load_parameters(parameters if parameters is not None else [])

    @property
    def _path(self):
        if self._parent is None:
            return []
        return self._parent._path + [self._name]

    def _get_path_val_param(self, absolute=True):
        raise NotImplementedError()

    def _set_to_path(self, path, value, overwrite=False):
        """Will set the value to the provided path. Local node if path length
        is one, a child node if path length is more that one.

        path: list
        value: the value to set
        """

    def _load_parameters(self, parameters):
        for parameter in parameters:
            node_type = parameter.get("type", "dict")
            node_type = TYPE_MAPPING.get(node_type) or node_type
            name = next(key for key in parameter if key != "type")
            is_node = isinstance(parameter[name], list)
            if node_type is dict:
                if is_node and not self._has_attr(name):
                    node = ConfNode(
                        parameters=parameter[name], name=name, parent=self
                    )
                    setattr(self, name, node)
                elif is_node:
                    getattr(self, name)._load_parameters(parameter[name])
                else:
                    self._load_parameter(name, parameter[name])
            elif node_type is list:
                if not self._has_attr(name):
                    node = ListNode(
                        parameters=parameter[name] if is_node else None,
                        node_type=None if is_node else parameter[name],
                        name=name,
                        parent=self,
                    )
                    setattr(self, name, node)
                else:
                    raise Exception("_load_parameters")
            if name not in self._children:
                self._children.append(name)

    def _load_parameter(self, name, settings, node_type=dict):
        if name in self._parameters:
            logger.debug("ignoring")
            return
        has_default = bool(settings.get("default"))
        has_type = bool(settings.get("type"))
        # something smarter that'd allow custom type
        if has_default and not has_type:
            settings["type"] = type(settings["default"])
        else:
            if settings.get("type") in TYPE_MAPPING:
                settings["type"] = TYPE_MAPPING[settings["type"]]
            elif isinstance(settings.get("type"), type):
                pass
            elif has_type:
                logger.warning("unknown type %r", settings["type"])
                settings["type"] = str
            else:
                settings["type"] = str
        has_among = bool(settings.get("among"))
        settings["required"] = bool(settings.get("required"))
        settings["read_only"] = bool(settings.get("read_only"))

        path = ".".join(map(str, chain(self._path, [name])))
        if has_among:
            assert isinstance(settings["among"], list), (
                f"parameters {path!r} configuration has wrong value for "
                "'among', should be a list, ignoring it"
            )
        if has_default and has_among:
            assert settings.get("default") in settings.get("among"), (
                f"default value for {path!r} is not among the "
                f"selectable values ({settings.get('among')!r}"
            )
        if has_default and settings["required"]:
            raise ValueError(
                f"{path!r} required parameter can't have default value"
            )

        if "type" in settings and "default" in settings:
            settings["default"] = settings["type"](settings["default"])
        self._parameters[name] = settings

    def _has_attr(self, attr):
        try:
            super().__getattribute__(attr)
            return True
        except AttributeError:
            return False


class ConfNode(AbstractNode):
    def _set_to_path(self, path, value, overwrite=False):
        attr = path[0]
        if len(path) == 1:
            if not overwrite and self._has_attr(attr):
                return
            read_only = None
            if self._parameters[attr].get("read_only"):
                # bypassing overwrite protection since we either:
                # - do not have that value in the first place
                # - overwrite has been activated
                read_only = self._parameters[attr].pop("read_only")
            res = setattr(self, attr, value)
            if read_only is not None:
                self._parameters[attr]["read_only"] = read_only
            return res
        return getattr(self, attr)._set_to_path(
            path[1:], value, overwrite=overwrite
        )

    def __getattribute__(self, name):
        """Return a parameter of the node if this one is defined.
        Its default value if it has one.
        """
        if name.startswith("_"):
            return super().__getattribute__(name)
        if "default" in self._parameters.get(name, {}):
            try:  # Trying to get attr, if AttributeError => is absent
                return super().__getattribute__(name)
            except AttributeError:
                return self._parameters[name]["default"]
        return super().__getattribute__(name)

    def __setattr__(self, key, value):
        if key.startswith("_") or isinstance(value, AbstractNode):
            return super().__setattr__(key, value)
        if key not in self._parameters:
            raise ValueError(f"{self._path} is not a registered conf option")
        if self._parameters[key].get("read_only"):
            raise AttributeError("attribute is in read only mode")
        if "among" in self._parameters[key]:
            if value not in self._parameters[key]["among"]:
                raise ValueError(
                    f"{self._path!r}: value {value!r} isn't "
                    f"in {self._parameters[key]['among']!r}"
                )
        if "type" in self._parameters[key]:
            value = self._parameters[key]["type"](value)
        return super().__setattr__(key, value)

    def _get_path_val_param(self, absolute=True):
        for child in self._children:
            if isinstance(getattr(self, child, None), AbstractNode):
                yield from getattr(self, child)._get_path_val_param()
            else:
                if absolute:
                    path = self._path + [child]
                else:
                    path = [child]
                yield path, getattr(self, child, NoValue), self._parameters[
                    child
                ]

    def __repr__(self):
        result = {"string": f"<{self.__class__.__name__}({{"}
        result["length"] = len(result["string"])
        old_loc, new_loc = [], []

        def spaces(index):
            if len(result["string"]) != result["length"]:
                return " " * (4 * index + result["length"])
            return ""

        def open_path(index, name):
            result["string"] += f"{spaces(index)}{name!r}: {{\n"

        def close_path(index):
            result["string"] += f"{spaces(index - 1)}}},\n"

        def add_key(index, name, value):
            result["string"] += f"{spaces(index)}{name!r}: {value!r},\n"

        for path, value, _ in self._get_path_val_param():
            new_loc = path[:-1]
            if new_loc != old_loc:
                diff_index = None
                for index, old_new in enumerate(zip(old_loc, new_loc)):
                    if old_new[0] != old_new[1]:
                        diff_index = index
                        break
                if diff_index is not None:
                    for index in range(len(old_loc), diff_index, -1):
                        close_path(index)
                    for index in range(diff_index, len(new_loc)):
                        open_path(index, new_loc[index])
                elif len(old_loc) > len(new_loc):  # we got out
                    for index in range(len(old_loc), len(new_loc), -1):
                        close_path(index)
                elif len(new_loc) > len(old_loc):  # we got in
                    for index in range(len(old_loc), len(new_loc)):
                        open_path(index, new_loc[index])
            add_key(len(new_loc), path[-1], value)
            old_loc = new_loc
        for index in range(len(old_loc), 0, -1):
            close_path(index)
        return result["string"] + ")>"


class ListNode(list, AbstractNode):
    def __init__(
        self,
        *args,
        parameters=None,
        parent=None,
        name="",
        node_type=None,
        **kwargs,
    ):
        AbstractNode.__init__(self, parameters, parent, name)
        list.__init__(self, *args, **kwargs)
        self._node_type = node_type or {}
        if self._children:
            parameters = [
                {child: self._parameters[child]} for child in self._children
            ]
            self._template_node = ConfNode(parameters, parent=self, name=Index)

    def _get_path_val_param(self, absolute=True):
        path = self._path
        if self._children:
            for child in self._children:
                if self:
                    for index, value in enumerate(self):
                        yield from value._get_path_val_param()
                else:
                    yield (
                        path + [Index, child],
                        NoValue,
                        self._parameters[child],
                    )
        else:
            if self:
                for index, value in enumerate(self):
                    yield path + [index], value, self._parameters
            else:
                yield path + [Index], self, self._parameters

    def __setitem__(self, index, value):
        if self._node_type.get("type"):
            if not isinstance(value, self._node_type["type"]):
                value = self._node_type["type"](value)
            return super().__setitem__(index, value)
        node = deepcopy(self._template_node)
        for path, _, _ in node._get_path_val_param(absolute=False):
            for _, sub_value in extract_value(value, path):
                node._set_to_path(path, sub_value)
        return super().__setitem__(index, node)

    def _set_to_path(self, path, value, overwrite=False):
        assert isinstance(path[0], int)
        if len(path) == 1 and not self._children:
            if len(self) <= path[0]:
                self.append(value)
            else:
                self[path[0]] = value
        else:
            if len(self) <= path[0]:
                node = deepcopy(self._template_node)
                self.append(node)
            self[path[0]]._set_to_path(path[1:], value, overwrite)



import json
import logging
from os.path import abspath, expanduser, splitext

import yaml

from the_conf.utils import Index

logger = logging.getLogger(__name__)


def read(*paths):
    any_found = False
    for path in paths:
        path = abspath(expanduser(path.strip()))
        ext = splitext(path)[1][1:]
        try:
            if ext in {"yml", "yaml"}:
                with open(path, "r", encoding="utf8") as fd:
                    yield path, ext, yaml.load(
                        fd.read(), Loader=yaml.FullLoader
                    )
            elif ext == "json":
                with open(path, "r", encoding="utf8") as fd:
                    yield path, ext, json.load(fd)
            else:
                logger.error("File %r ignored: unknown type (%s)", path, ext)
                continue
            any_found = True
        except FileNotFoundError:
            logger.debug("%r not found", path)
        except PermissionError:
            logger.warning("%r: no right to read", path)
    if not any_found:
        logger.warning("no file found among %r", paths)


def extract_value(config, path, full_path=None):
    full_path = full_path or []
    if len(path) == 1 and path[0] in config:
        yield full_path + [path[0]], config[path[0]]
    elif path[0] is Index and isinstance(config, (list, tuple)):
        for index, sub_config in enumerate(config):
            if len(path) == 1:
                yield full_path + [index], sub_config
            else:
                yield from extract_value(
                    config[index], path[1:], full_path + [index]
                )
    elif path[0] in config:
        yield from extract_value(
            config[path[0]], path[1:], full_path + [path[0]]
        )
    else:
        raise ValueError(f"no {path[0]!r} in {config!r}")


def extract_values(paths, config, config_file):
    for path in paths:
        try:
            if Index in path:
                for full_path, sub_value in extract_value(config, path):
                    yield full_path, sub_value
            else:
                for full_path, value in extract_value(config, path):
                    assert full_path == path
                    yield path, value
        except ValueError:
            logger.debug("%s not found in %r", ".".join(path), config_file)


def write(config, path):
    path = abspath(expanduser(path.strip()))
    ext = splitext(path)[1][1:]
    if ext in {"yml", "yaml"}:
        with open(path, "w", encoding="utf8") as fp:
            yaml.dump(config, fp)
    elif ext == "json":
        with open(path, "w", encoding="utf8") as fp:
            json.dump(config, fp)
    else:
        raise ValueError(
            "couldn't make out file type, conf file path should "
            "end with either yml, yaml or json"
        )



from argparse import ArgumentParser

CONFIG_OPT_DEST = "config_file_path"


def path_to_cmd_opt(path):
    return "--" + "-".join(map(str.lower, path))


def path_to_dest(path):
    return "_".join(path)


def get_parser(path_n_params, config_file_cmd_line):
    parser = ArgumentParser()
    parser.add_argument(
        *config_file_cmd_line,
        dest=CONFIG_OPT_DEST,
        help="set main conf file to load configuration from",
    )
    for path, _, param in path_n_params:
        parser_kw = {}

        if param.get("no_cmd"):
            continue
        flag = (
            param["cmd_line_opt"]
            if param.get("cmd_line_opt")
            else path_to_cmd_opt(path)
        )

        if "type" in param:
            parser_kw["type"] = param["type"]
            if param["type"] is bool and param.get("default") is False:
                param["action"], param["default"] = "store_true", False
            elif param["type"] is bool and param.get("default") is True:
                param["action"], param["default"] = "store_false", True
        if "among" in param:
            parser_kw["choices"] = param["among"]
        if "help_txt" in param:
            parser_kw["help"] = param["help_txt"]

        parser.add_argument(flag, dest=path_to_dest(path), **parser_kw)
    return parser


def yield_values_from_cmd(path_val_params, opts, config_file_cmd_line):
    parser = get_parser(path_val_params, config_file_cmd_line)
    cmd_line_args, _ = parser.parse_known_args(opts)
    yield getattr(cmd_line_args, CONFIG_OPT_DEST)
    for path, _, _ in path_val_params:
        value = getattr(cmd_line_args, path_to_dest(path))
        if value is not None:
            yield path, value



import logging
import os

from the_conf import command_line, files, interractive, node, utils

logger = logging.getLogger(__name__)
DEFAULT_ORDER = "cmd", "files", "env"
DEFAULT_CONFIG_FILE_CMD_LINE = "-C", "--config"
DEFAULT_CONFIG_FILE_ENVIRON = ("CONFIG_FILE",)


class TheConf(node.ConfNode):
    def __init__(
        self, *metaconfs, prompt_values=False, cmd_line_opts=None, environ=None
    ):
        self._source_order = list(DEFAULT_ORDER)
        self._config_files = []
        self._config_file_cmd_line = list(DEFAULT_CONFIG_FILE_CMD_LINE)
        self._config_file_environ = list(DEFAULT_CONFIG_FILE_ENVIRON)
        self._main_conf_file = None
        self._cmd_line_opts = cmd_line_opts
        self._environ = environ
        self._prompt_values = prompt_values

        def is_default(value, default):
            if not value or isinstance(value, tuple):
                return True
            return tuple(value) == default

        def set_metaconf_setting(key, metaconf, default):
            if key not in metaconf:
                return
            new_value = metaconf[key]
            if isinstance(metaconf[key], (list, tuple, set)):
                new_value = list(new_value)
            elif isinstance(new_value, (str, int, float)):
                raise TypeError(
                    f"metaconf parameter {key} is "
                    f"of unknown type {type(new_value)!r}"
                )
            value = getattr(self, "_" + key)
            if is_default(value, default):
                setattr(self, "_" + key, new_value)
            else:
                value.extend(new_value)

        super().__init__()
        for mc in metaconfs:
            if isinstance(mc, str):
                _, _, mc = next(files.read(mc))
            set_metaconf_setting("source_order", mc, DEFAULT_ORDER)
            set_metaconf_setting(
                "config_file_cmd_line", mc, DEFAULT_CONFIG_FILE_CMD_LINE
            )
            set_metaconf_setting(
                "config_file_environ", mc, DEFAULT_CONFIG_FILE_ENVIRON
            )
            set_metaconf_setting("config_files", mc, None)

            self._load_parameters(mc["parameters"])
        self.load()

    def _load_files(self):
        if self._config_files is None:
            return
        for conf_file, _, config in files.read(*self._config_files):
            paths = list(path for path, _, _ in self._get_path_val_param())
            for path, value in files.extract_values(paths, config, conf_file):
                self._set_to_path(path, value, overwrite=True)

    def _load_cmd(self, opts=None):
        gen = command_line.yield_values_from_cmd(
            list(self._get_path_val_param()),
            self._cmd_line_opts,
            self._config_file_cmd_line,
        )
        config_file = next(gen)
        if config_file:
            self._config_files.insert(0, config_file)

        for path, value in gen:
            self._set_to_path(path, value, overwrite=True)

    def _load_env(self, environ=None):
        if environ is None:
            environ = os.environ
        for config_env_key in self._config_file_environ:
            if config_env_key in environ:
                self._config_files.insert(0, environ[config_env_key])
        for path, _, _ in self._get_path_val_param():
            env_key = "_".join(map(str.upper, path))
            if env_key in environ:
                self._set_to_path(path, environ[env_key], overwrite=True)

    def load(self):
        for order in self._source_order:
            if order == "files":
                self._load_files()
            elif order == "cmd":
                self._load_cmd(self._cmd_line_opts)
            elif order == "env":
                self._load_env(self._environ)
            else:
                raise Exception(f"unknown order {order!r}")

        if self._prompt_values:
            self.prompt_values(False, False, False, False)

        for path, value, param in self._get_path_val_param():
            if value is utils.NoValue and param.get("required"):
                raise ValueError(
                    f"loading finished and {'.'.join(path)!r} " "is not set"
                )

    def _extract_config(self):
        config = {}
        for paths, value, param in self._get_path_val_param():
            if value is utils.NoValue:
                continue
            if "default" in param and value == param["default"]:
                continue
            curr_config = config
            for path in paths[:-1]:
                curr_config[path] = {}
                curr_config = curr_config[path]
            curr_config[paths[-1]] = value
        return config

    def write(self, config_file=None):
        if config_file is None and not self._config_files:
            raise ValueError("no config file to write in")

        files.write(
            self._extract_config(), config_file or self._config_files[0]
        )

    def prompt_values(
        self,
        only_empty=True,
        only_no_default=True,
        only_required=True,
        only_w_help=True,
    ):
        for path, value, param in self._get_path_val_param():
            if only_w_help and not param.get("help_txt"):
                continue
            if only_required and not param.get("required"):
                continue
            if only_no_default and not param.get("default"):
                continue
            if only_empty and value is not utils.NoValue:
                continue
            if param.get("type") is bool:
                self._set_to_path(
                    path,
                    interractive.ask_bool(
                        param.get("help_txt", ".".join(path)),
                        default=param.get("default"),
                        required=param.get("required"),
                    ),
                )
            else:
                self._set_to_path(
                    path,
                    interractive.ask(
                        param.get("help_txt", ".".join(path)),
                        choices=param.get("among"),
                        default=param.get("default"),
                        required=param.get("required"),
                        cast=param.get("type"),
                    ),
                )
